<?php

require_once PROJECT_ROOT_PATH . "/src/db/DBConnection.php";

abstract class BaseRepository
{
    protected static $instance;
    protected $pdo;
    protected $tableName = '';

    protected function __construct()
    {
        $this->pdo = DBConnection::getInstance()->getPDO();
    }

    public static function getInstance()
    {
        if (!isset(self::$instance)) {
            self::$instance = new static();
        }
        return self::$instance;
    }

    public function getPDO()
    {
        return $this->pdo;
    }

    public function insert($data, $paramTypes)
    {

        // Filter the $data array to include only the keys present in $paramTypes
        $filteredData = array_intersect_key($data, $paramTypes);

        $columns = implode(', ', array_keys($filteredData));
        $placeholders = ':' . implode(', :', array_keys($filteredData));

        $sql = "INSERT INTO $this->tableName ($columns) VALUES ($placeholders)";

        $stmt = $this->pdo->prepare($sql);

        foreach ($filteredData as $key => $value) {
            $stmt->bindValue(':' . $key, $value, $paramTypes[$key]);
        }

        if ($stmt->execute()) {
            return $this->pdo->lastInsertId();
        } else {
            return false;
        }
    }

    public function selectOne($columns = "*", $where = [])
    {
        $query = "SELECT $columns FROM $this->tableName";

        if (!empty($where)) {
            $query .= " WHERE " . implode(" AND ", $where);
        }
        $query .= ";";

        $stmt = $this->pdo->prepare($query);
        $stmt->execute();

        return $stmt->fetch();
    }

    public function select($columns = "*", $where = [], $orderBy = [], $page = 1, $perPage = 10, $isDesc = false)
    {
        $query = "SELECT $columns FROM $this->tableName";

        if (!empty($where)) {
            $query .= " WHERE " . implode(" AND ", $where);
        }

        if (!empty($orderBy)) {
            $orderByWithDesc = array_map(function ($column) use ($isDesc) {
                return $column . ($isDesc && $isDesc != "false" ? " DESC" : "");
            }, $orderBy);
            $query .= " ORDER BY " . implode(", ", $orderByWithDesc);
        }

        if (!empty($perPage)) {
            $offset = ($page - 1) * $perPage;
            $query .= " LIMIT $perPage OFFSET $offset";
        }
        // UNCOMMENT THIS FOR DEBUG
        // echo $query;
        $stmt = $this->pdo->prepare($query);
        $stmt->execute();

        $results = $stmt->fetchAll(PDO::FETCH_ASSOC);


        $countQuery = "SELECT COUNT(*) as total FROM $this->tableName";
        if (!empty($where)) {
            $countQuery .= " WHERE " . implode(" AND ", $where);
        }

        $stmtCount = $this->pdo->prepare($countQuery);
        $stmtCount->execute();
        $totalCount = $stmtCount->fetch(PDO::FETCH_ASSOC)['total'];

        return [
            'count' => $totalCount,
            'result' => $results,
        ];
    }

    public function update($id_col, $id, $data, $paramTypes)
    {
        $filteredData = array_intersect_key($data, $paramTypes);

        $setClause = implode(', ', array_map(function ($key) {
            return "$key = :$key";
        }, array_keys($filteredData)));

        $sql = "UPDATE $this->tableName SET $setClause WHERE $id_col = $id";

        $stmt = $this->pdo->prepare($sql);
        foreach ($filteredData as $key => $value) {
            $stmt->bindValue(':' . $key, $value, $paramTypes[$key]);
        }

        return $stmt->execute();
    }

    public function delete($id_col, $id)
    {
        $sql = "DELETE FROM $this->tableName WHERE $id_col = $id";

        $stmt = $this->pdo->prepare($sql);

        return $stmt->execute();
    }

    public function count()
    {
        $query = "SELECT count(*) FROM $this->tableName";
        $stmt = $this->pdo->prepare($query);
        $stmt->execute();

        return $stmt->fetch();

    }
}